@@ -1,4 +1,5 @@
 #include <pybind11/pybind11.h>
+#include <pybind11/operators.h>
 
 #include <osmium/osm.hpp>
 #include <osmium/osm/entity_bits.hpp>
@@ -53,6 +54,7 @@ PYBIND11_MODULE(_osm, m) {
          "necessarily have to be valid.")
         .def(py::init<>())
         .def(py::init<double, double>())
+        .def(py::self == py::self)
         .def_property_readonly("x", &osmium::Location::x,
              "(read-only) X coordinate (longitude) as a fixed-point integer.")
         .def_property_readonly("y", &osmium::Location::y,
@@ -100,7 +102,8 @@ PYBIND11_MODULE(_osm, m) {
              py::return_value_policy::reference_internal,
              "Extend the box to include the given location. If the location "
              "is invalid the box remains unchanged. If the box is invalid, it "
-             "will contain only the location after the operation.")
+             "will contain only the location after the operation. "
+             "Returns a reference to itself.")
         .def("extend",
              (osmium::Box& (osmium::Box::*)(osmium::Box const &))
                  &osmium::Box::extend,
@@ -108,7 +111,8 @@ PYBIND11_MODULE(_osm, m) {
              py::return_value_policy::reference_internal,
              "Extend the box to include the given box. If the box to be added "
              "is invalid the input box remains unchanged. If the input box is invalid, it "
-             "will become equal to the box that was added.")
+             "will become equal to the box that was added. "
+             "Returns a reference to itself.")
         .def("valid", &osmium::Box::valid,
              "Check if the box coordinates are defined and with the usual bounds.")
         .def("size", &osmium::Box::size,
@@ -207,11 +211,11 @@ PYBIND11_MODULE(_osm, m) {
         "an immutable sequence of :py:class:`osmium.osm.NodeRef`. This class "
         "is normally not used directly, use one of its subclasses instead.")
         .def("__len__", &osmium::NodeRefList::size)
-        .def("__getitem__", [](osmium::NodeRefList const &list, ssize_t idx)
+        .def("__getitem__", [](osmium::NodeRefList const &list, Py_ssize_t idx)
              {
                 auto sz = list.size();
                 osmium::NodeRefList::size_type iout =
-                    (idx >= 0 ? idx : (ssize_t) sz + idx);
+                    (idx >= 0 ? idx : (Py_ssize_t) sz + idx);
 
                 if (iout >= sz || iout < 0) {
                     throw py::index_error("Bad index.");
