@@ -1,12 +1,15 @@
 """ Helper functions to communicate with replication servers.
 """
-
+from typing import NamedTuple, Optional, Any, Iterator, cast, Mapping, Tuple
 import requests
 import urllib.request as urlrequest
+from urllib.error import URLError
 import datetime as dt
 from collections import namedtuple
+from contextlib import contextmanager
 from math import ceil
-from osmium import MergeInputReader
+
+from osmium import MergeInputReader, BaseHandler
 from osmium import io as oio
 from osmium import version
 
@@ -15,47 +18,85 @@ import logging
 LOG = logging.getLogger('pyosmium')
 LOG.addHandler(logging.NullHandler())
 
-OsmosisState = namedtuple('OsmosisState', ['sequence', 'timestamp'])
-DownloadResult = namedtuple('DownloadResult', ['id', 'reader', 'newest'])
+class OsmosisState(NamedTuple):
+    sequence: int
+    timestamp: dt.datetime
+
+class DownloadResult(NamedTuple):
+    id: int
+    reader: MergeInputReader
+    newest: int
 
 class ReplicationServer:
-    """ Represents a server that publishes replication data. Replication
-        change files allow to keep local OSM data up-to-date without downloading
-        the full dataset again.
+    """ Represents a connection to a  server that publishes replication data.
+        Replication change files allow to keep local OSM data up-to-date without
+        downloading the full dataset again.
+
+        `url` contains the base URL of the replication service. This is the
+        directory that contains the state file with the current state. If the
+        replication service serves something other than osc.gz files, set
+        the `diff_type` to the given file suffix.
+
+        ReplicationServer may be used as a context manager. In this case, it
+        internally keeps a connection to the server making downloads faster.
     """
 
-    def __init__(self, url, diff_type='osc.gz'):
+    def __init__(self, url: str, diff_type: str = 'osc.gz') -> None:
         self.baseurl = url
         self.diff_type = diff_type
-        self.session = requests.Session()
+        self.extra_request_params: Mapping[str, Any] = dict(timeout=60, stream=True)
+        self.session: Optional[requests.Session] = None
 
-    def make_request(self, url):
-        headers = {"User-Agent" : "pyosmium/{}".format(version.pyosmium_release)}
+    def close(self) -> None:
+        """ Close any open connection to the replication server.
+        """
+        if self.session is not None:
+            self.session.close()
+            self.session = None
+
+    def __enter__(self) -> 'ReplicationServer':
+        self.session = requests.Session()
+        return self
+
+    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
+        self.close()
+
+    def set_request_parameter(self, key: str, value: Any) -> None:
+        """ Set a parameter which will be handed to the requests library
+            when calling `requests.get()`. This
+            may be used to set custom headers, timeouts and similar parameters.
+            See the `requests documentation <https://requests.readthedocs.io/en/latest/api/?highlight=get#requests.request>`_
+            for possible parameters. Per default, a timeout of 60 sec is set
+            and streaming download enabled.
+        """
+        self.extra_request_params[key] = value
+
+    def make_request(self, url: str) -> urlrequest.Request:
+        headers = {"User-Agent" : f"pyosmium/{version.pyosmium_release}"}
         return urlrequest.Request(url, headers=headers)
 
-    def open_url(self, url):
+    def open_url(self, url: urlrequest.Request) -> Any:
         """ Download a resource from the given URL and return a byte sequence
             of the content.
-
-            This method has no support for cookies or any special authentication
-            methods. If you need these, you have to provide your own custom URL
-            opener. The method has to return an object which supports the
-            `read()` and `readline()` methods to access the content. Example::
-
-                def my_open_url(self, url):
-                    opener = urlrequest.build_opener()
-                    opener.addheaders = [('X-Fancy-Header', 'important_content')]
-                    return opener.open(url)
-
-                svr = ReplicationServer()
-                svr.open_url = my_open_url
         """
-        headers = dict()
-        for h in url.header_items():
-            headers[h[0]] = h[1]
-        return self.session.get(url.get_full_url(), headers=headers, stream=True)
+        if 'headers' in self.extra_request_params:
+            get_params = self.extra_request_params
+        else:
+            get_params = dict(self.extra_request_params)
+            get_params['headers'] = {k: v for k,v in url.header_items()}
 
-    def collect_diffs(self, start_id, max_size=1024):
+        if self.session is not None:
+            return self.session.get(url.get_full_url(), **get_params)
+
+        @contextmanager
+        def _get_url_with_session() -> Iterator[requests.Response]:
+            with requests.Session() as session:
+                request = session.get(url.get_full_url(), **get_params)
+                yield request
+
+        return _get_url_with_session()
+
+    def collect_diffs(self, start_id: int, max_size: int = 1024) -> Optional[DownloadResult]:
         """ Create a MergeInputReader and download diffs starting with sequence
             id `start_id` into it. `max_size`
             restricts the number of diffs that are downloaded. The download
@@ -100,7 +141,9 @@ class ReplicationServer:
 
         return DownloadResult(current_id - 1, rd, newest.sequence)
 
-    def apply_diffs(self, handler, start_id, max_size=1024, idx="", simplify=True):
+    def apply_diffs(self, handler: BaseHandler, start_id: int,
+                    max_size: int = 1024, idx: str = "",
+                    simplify: bool = True) -> Optional[int]:
         """ Download diffs starting with sequence id `start_id`, merge them
             together and then apply them to handler `handler`. `max_size`
             restricts the number of diffs that are downloaded. The download
@@ -134,9 +177,11 @@ class ReplicationServer:
 
         return diffs.id
 
-    def apply_diffs_to_file(self, infile, outfile, start_id, max_size=1024,
-                            set_replication_header=True, extra_headers=None,
-                            outformat=None):
+    def apply_diffs_to_file(self, infile: str, outfile: str,
+                            start_id: int, max_size: int = 1024,
+                            set_replication_header: bool = True,
+                            extra_headers: Optional[Mapping[str, str]] = None,
+                            outformat: Optional[str] = None) -> Optional[Tuple[int, int]]:
         """ Download diffs starting with sequence id `start_id`, merge them
             with the data from the OSM file named `infile` and write the result
             into a file with the name `outfile`. The output file must not yet
@@ -199,7 +244,8 @@ class ReplicationServer:
         return (diffs.id, diffs.newest)
 
 
-    def timestamp_to_sequence(self, timestamp, balanced_search=False):
+    def timestamp_to_sequence(self, timestamp: dt.datetime,
+                              balanced_search: bool = False) -> Optional[int]:
         """ Get the sequence number of the replication file that contains the
             given timestamp. The search algorithm is optimised for replication
             servers that publish updates in regular intervals. For servers
@@ -281,7 +327,7 @@ class ReplicationServer:
                 return lower.sequence
 
 
-    def get_state_info(self, seq=None, retries=2):
+    def get_state_info(self, seq: Optional[int] = None, retries: int = 2) -> Optional[OsmosisState]:
         """ Downloads and returns the state information for the given
             sequence. If the download is successful, a namedtuple with
             `sequence` and `timestamp` is returned, otherwise the function
@@ -289,61 +335,60 @@ class ReplicationServer:
             is retried when pyosmium detects a truncated state file.
         """
         for _ in range(retries + 1):
+            ts = None
+            next_seq = None
+
             try:
-                response = self.open_url(self.make_request(self.get_state_url(seq)))
-            except Exception as err:
+                with self.open_url(self.make_request(self.get_state_url(seq))) as response:
+                    if hasattr(response, 'iter_lines'):
+                        # generated by requests
+                        lines = response.iter_lines()
+                    else:
+                        lines = response
+
+                    for line in lines:
+                        line = line.decode('utf-8').strip()
+                        if '#' in line:
+                            line = line[0:line.index('#')]
+                        else:
+                            line = line.strip()
+                        if line:
+                            kv = line.split('=', 2)
+                            if len(kv) != 2:
+                                return None
+                            if kv[0] == 'sequenceNumber':
+                                next_seq = int(kv[1])
+                            elif kv[0] == 'timestamp':
+                                try:
+                                    ts = dt.datetime.strptime(kv[1], "%Y-%m-%dT%H\\:%M\\:%SZ")
+                                except ValueError:
+                                    break
+                                ts = ts.replace(tzinfo=dt.timezone.utc)
+
+            except (URLError, IOError) as err:
                 LOG.debug("Loading state info %s failed with: %s", seq, str(err))
                 return None
 
-            ts = None
-            seq = None
-            if hasattr(response, 'iter_lines'):
-                # generated by requests
-                get_line = response.iter_lines
-            else:
-                # generated by urllib.request
-                get_line = response.readline
-
-            for line in get_line():
-                line = line.decode('utf-8')
-                if '#' in line:
-                    line = line[0:line.index('#')]
-                else:
-                    line = line.strip()
-                if line:
-                    kv = line.split('=', 2)
-                    if len(kv) != 2:
-                        return None
-                    if kv[0] == 'sequenceNumber':
-                        seq = int(kv[1])
-                    elif kv[0] == 'timestamp':
-                        try:
-                            ts = dt.datetime.strptime(kv[1], "%Y-%m-%dT%H\\:%M\\:%SZ")
-                        except ValueError:
-                            break
-                        ts = ts.replace(tzinfo=dt.timezone.utc)
-
-            if ts is not None and seq is not None:
-                return OsmosisState(sequence=seq, timestamp=ts)
+            if ts is not None and next_seq is not None:
+                return OsmosisState(sequence=next_seq, timestamp=ts)
 
         return None
 
-    def get_diff_block(self, seq):
+    def get_diff_block(self, seq: int) -> str:
         """ Downloads the diff with the given sequence number and returns
             it as a byte sequence. Throws a :code:`urllib.error.HTTPError`
-            (or :code:`urllib2.HTTPError` in python2)
             if the file cannot be downloaded.
         """
-        resp = self.open_url(self.make_request(self.get_diff_url(seq)))
-        if hasattr(resp, 'content'):
-            # generated by requests
-            return resp.content
-        else:
+        with self.open_url(self.make_request(self.get_diff_url(seq))) as resp:
+            if hasattr(resp, 'content'):
+                # generated by requests
+                return cast(str, resp.content)
+
             # generated by urllib.request
-            return resp.read()
+            return cast(str, resp.read())
 
 
-    def get_state_url(self, seq):
+    def get_state_url(self, seq: Optional[int]) -> str:
         """ Returns the URL of the state.txt files for a given sequence id.
 
             If seq is `None` the URL for the latest state info is returned,
@@ -357,7 +402,7 @@ class ReplicationServer:
                (self.baseurl, seq / 1000000, (seq % 1000000) / 1000, seq % 1000)
 
 
-    def get_diff_url(self, seq):
+    def get_diff_url(self, seq: int) -> str:
         """ Returns the URL to the diff file for the given sequence id.
         """
         return '%s/%03i/%03i/%03i.%s' % \
