@@ -1,17 +1,20 @@
 """ Helper functions for change file handling. """
-
+from typing import NamedTuple, Optional
 import logging
 import datetime as dt
 from collections import namedtuple
 from osmium.io import Reader as oreader
 from osmium.osm import NOTHING
 
-log = logging.getLogger('pyosmium')
+LOG = logging.getLogger('pyosmium')
 
-ReplicationHeader = namedtuple('ReplicationHeader',
-                                ['url', 'sequence', 'timestamp'])
+class ReplicationHeader(NamedTuple):
+    url: Optional[str]
+    sequence: Optional[int]
+    timestamp: Optional[dt.datetime]
 
-def get_replication_header(fname):
+
+def get_replication_header(fname: str) -> ReplicationHeader:
     """ Scans the given file for an Osmosis replication header. It returns
         a namedtuple with `url`, `sequence` and `timestamp`. Each or all fields
         may be None, if the piece of information is not avilable. If any of
@@ -24,25 +27,26 @@ def get_replication_header(fname):
     r = oreader(fname, NOTHING)
     h = r.header()
 
-    ts = h.get("osmosis_replication_timestamp")
-    url = h.get("osmosis_replication_base_url")
+    tsstr = h.get("osmosis_replication_timestamp")
+    url: Optional[str] = h.get("osmosis_replication_base_url")
 
-    if url or ts:
-        log.debug("Replication information found in OSM file header.")
+    if url or tsstr:
+        LOG.debug("Replication information found in OSM file header.")
 
     if url:
-        log.debug("Replication URL: %s" % url)
+        LOG.debug("Replication URL: %s", url)
         # the sequence ID is only considered valid, if an URL is given
-        seq = h.get("osmosis_replication_sequence_number")
-        if seq:
-            log.debug("Replication sequence: %s" % seq)
+        seqstr = h.get("osmosis_replication_sequence_number")
+        seq: Optional[int]
+        if seqstr:
+            LOG.debug("Replication sequence: %s", seqstr)
             try:
-                seq = int(seq)
+                seq = int(seqstr)
                 if seq < 0:
-                    log.warning("Sequence id '%d' in OSM file header is negative. Ignored." % seq)
+                    LOG.warning("Sequence id '%d' in OSM file header is negative. Ignored.", seq)
                     seq = None
             except ValueError:
-                log.warning("Sequence id '%s' in OSM file header is not a number.Ignored" % seq)
+                LOG.warning("Sequence id '%s' in OSM file header is not a number. Ignored.", seq)
                 seq = None
         else:
             seq = None
@@ -50,14 +54,14 @@ def get_replication_header(fname):
         url = None
         seq = None
 
-    if ts:
-        log.debug("Replication timestamp: %s" % ts)
+    if tsstr:
+        LOG.debug("Replication timestamp: %s", tsstr)
         try:
-            ts = dt.datetime.strptime(ts, "%Y-%m-%dT%H:%M:%SZ")
+            ts = dt.datetime.strptime(tsstr, "%Y-%m-%dT%H:%M:%SZ")
             ts = ts.replace(tzinfo=dt.timezone.utc)
 
         except ValueError:
-            log.warning("Date in OSM file header is not in ISO8601 format (e.g. 2015-12-24T08:08Z). Ignored")
+            LOG.warning("Date in OSM file header is not in ISO8601 format (e.g. 2015-12-24T08:08Z). Ignored.")
             ts = None
     else:
         ts = None
